@@ -466,7 +466,9 @@ int spa_alsa_init(struct state *state, const struct spa_dict *info)
 	}
 	if (state->clock_name[0] == '\0')
 		snprintf(state->clock_name, sizeof(state->clock_name),
-				"api.alsa.%u", state->card_index);
+				"api.alsa.%s-%u",
+				state->stream == SND_PCM_STREAM_PLAYBACK ? "p" : "c",
+				state->card_index);
 
 	if (state->stream == SND_PCM_STREAM_PLAYBACK) {
 		state->is_iec958 = spa_strstartswith(state->props.device, "iec958");
@@ -750,7 +752,7 @@ static bool uint32_array_contains(uint32_t *vals, uint32_t n_vals, uint32_t val)
 }
 
 static int add_rate(struct state *state, uint32_t scale, bool all, uint32_t index, uint32_t *next,
-		snd_pcm_hw_params_t *params, struct spa_pod_builder *b)
+		uint32_t min_allowed_rate, snd_pcm_hw_params_t *params, struct spa_pod_builder *b)
 {
 	struct spa_pod_frame f[1];
 	int err, dir;
@@ -761,6 +763,9 @@ static int add_rate(struct state *state, uint32_t scale, bool all, uint32_t inde
 	CHECK(snd_pcm_hw_params_get_rate_min(params, &min, &dir), "get_rate_min");
 	CHECK(snd_pcm_hw_params_get_rate_max(params, &max, &dir), "get_rate_max");
 
+	min_allowed_rate /= scale;
+	min = SPA_MAX(min_allowed_rate, min);
+
 	if (!state->multi_rate && state->card->format_ref > 0)
 		rate = state->card->rate;
 	else
@@ -1015,7 +1020,7 @@ static int enum_pcm_formats(struct state *state, uint32_t index, uint32_t *next,
 		choice->body.type = SPA_CHOICE_Enum;
 	spa_pod_builder_pop(b, &f[1]);
 
-	if ((res = add_rate(state, 1, false, index & 0xffff, next, params, b)) != 1)
+	if ((res = add_rate(state, 1, false, index & 0xffff, next, 0, params, b)) != 1)
 		return res;
 
 	if ((res = add_channels(state, false, index & 0xffff, next, params, b)) != 1)
@@ -1108,7 +1113,7 @@ static int enum_iec958_formats(struct state *state, uint32_t index, uint32_t *ne
 	}
 	spa_pod_builder_pop(b, &f[1]);
 
-	if ((res = add_rate(state, 1, true, index & 0xffff, next, params, b)) != 1)
+	if ((res = add_rate(state, 1, true, index & 0xffff, next, 0, params, b)) != 1)
 		return res;
 
 	(*next)++;
@@ -1163,7 +1168,14 @@ static int enum_dsd_formats(struct state *state, uint32_t index, uint32_t *next,
 	spa_pod_builder_prop(b, SPA_FORMAT_AUDIO_interleave, 0);
 	spa_pod_builder_int(b, interleave);
 
-	if ((res = add_rate(state, SPA_ABS(interleave), true, index & 0xffff, next, params, b)) != 1)
+	/* Use a lower rate limit of 352800 (= 44100 * 64 / 8). This is because in
+	 * PipeWire, DSD rates are given in bytes, not bits, so 352800 corresponds
+	 * to the bit rate of DSD64. (The "64" in DSD64 means "64 times the rate
+	 * of 44.1 kHz".) Some hardware may report rates lower than that, for example
+	 * 176400. This would correspond to "DSD32" (which does not exist). Trying
+	 * to use such a rate with DSD hardware does not work and may cause undefined
+	 * behavior in said hardware. */
+	if ((res = add_rate(state, SPA_ABS(interleave), true, index & 0xffff, next, 44100 * 64 / 8, params, b)) != 1)
 		return res;
 
 	if ((res = add_channels(state, true, index & 0xffff, next, params, b)) != 1)
@@ -1235,7 +1247,7 @@ spa_alsa_enum_format(struct state *state, int seq, uint32_t start, uint32_t num,
 
 int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_t flags)
 {
-	unsigned int rrate, rchannels, val;
+	unsigned int rrate, rchannels, val, rscale = 1;
 	snd_pcm_uframes_t period_size;
 	int err, dir;
 	snd_pcm_hw_params_t *params;
@@ -1322,21 +1334,26 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 		case 4:
 			rformat = SND_PCM_FORMAT_DSD_U32_BE;
 			rrate /= 4;
+			rscale = 4;
 			break;
 		case -4:
 			rformat = SND_PCM_FORMAT_DSD_U32_LE;
 			rrate /= 4;
+			rscale = 4;
 			break;
 		case 2:
 			rformat = SND_PCM_FORMAT_DSD_U16_BE;
 			rrate /= 2;
+			rscale = 2;
 			break;
 		case -2:
 			rformat = SND_PCM_FORMAT_DSD_U16_LE;
 			rrate /= 2;
+			rscale = 2;
 			break;
 		case 1:
 			rformat = SND_PCM_FORMAT_DSD_U8;
+			rscale = 1;
 			break;
 		default:
 			return -ENOTSUP;
@@ -1402,7 +1419,10 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 				state->props.device, rchannels, val);
 		if (!SPA_FLAG_IS_SET(flags, SPA_NODE_PARAM_FLAG_NEAREST))
 			return -EINVAL;
+		if (fmt->media_subtype != SPA_MEDIA_SUBTYPE_raw)
+			return -EINVAL;
 		rchannels = val;
+		fmt->info.raw.channels = rchannels;
 		match = false;
 	}
 
@@ -1422,7 +1442,10 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 				state->props.device, rrate, val);
 		if (!SPA_FLAG_IS_SET(flags, SPA_NODE_PARAM_FLAG_NEAREST))
 			return -EINVAL;
+		if (fmt->media_subtype != SPA_MEDIA_SUBTYPE_raw)
+			return -EINVAL;
 		rrate = val;
+		fmt->info.raw.rate = rrate;
 		match = false;
 	}
 
@@ -1430,6 +1453,7 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	state->channels = rchannels;
 	state->rate = rrate;
 	state->frame_size = snd_pcm_format_physical_width(rformat) / 8;
+	state->frame_scale = rscale;
 	state->planar = planar;
 	state->blocks = 1;
 	if (planar)
@@ -1477,7 +1501,7 @@ int spa_alsa_set_format(struct state *state, struct spa_audio_info *fmt, uint32_
 	} else {
 		CHECK(snd_pcm_hw_params_get_buffer_size_max(params, &state->buffer_frames), "get_buffer_size_max");
 
-		state->buffer_frames = SPA_MIN(state->buffer_frames, state->quantum_limit * 4);
+		state->buffer_frames = SPA_MIN(state->buffer_frames, state->quantum_limit * 4)* state->frame_scale;
 
 		CHECK(snd_pcm_hw_params_set_buffer_size_min(hndl, params, &state->buffer_frames), "set_buffer_size_min");
 		CHECK(snd_pcm_hw_params_set_buffer_size_near(hndl, params, &state->buffer_frames), "set_buffer_size_near");
@@ -1785,9 +1809,11 @@ static int update_time(struct state *state, uint64_t current_time, snd_pcm_sfram
 	if (err > state->max_error) {
 		err = state->max_error;
 		state->alsa_sync = true;
+		state->alsa_sync_warning = (diff == 0);
 	} else if (err < -state->max_error) {
 		err = -state->max_error;
 		state->alsa_sync = true;
+		state->alsa_sync_warning = (diff == 0);
 	}
 
 	if (!follower || state->matching)
@@ -1854,6 +1880,11 @@ static int setup_matching(struct state *state)
 		state->matching = false;
 
 	state->resample = ((uint32_t)state->rate != state->rate_denom) || state->matching;
+
+	spa_log_info(state->log, "driver clock:'%s'@%d our clock:'%s'@%d matching:%d resample:%d",
+			state->position->clock.name, state->rate_denom,
+			state->clock_name, state->rate,
+			state->matching, state->resample);
 	return 0;
 }
 
@@ -1867,6 +1898,7 @@ static inline void check_position_config(struct state *state)
 		state->duration = state->position->clock.duration;
 		state->rate_denom = state->position->clock.rate.denom;
 		state->threshold = (state->duration * state->rate + state->rate_denom-1) / state->rate_denom;
+		state->max_error = SPA_MAX(256.0f, state->threshold / 2.0f);
 		state->resample = ((uint32_t)state->rate != state->rate_denom) || state->matching;
 		state->alsa_sync = true;
 	}
@@ -1879,6 +1911,7 @@ int spa_alsa_write(struct state *state)
 	snd_pcm_uframes_t written, frames, offset, off, to_write, total_written, max_write;
 	snd_pcm_sframes_t commitres;
 	int res = 0;
+	size_t frame_size = state->frame_size;
 
 	check_position_config(state);
 
@@ -1893,9 +1926,18 @@ int spa_alsa_write(struct state *state)
 		if (SPA_UNLIKELY((res = get_status(state, current_time, &delay, &target)) < 0))
 			return res;
 
+		if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, true)) < 0))
+			return res;
+
 		if (SPA_UNLIKELY(state->alsa_sync)) {
-			spa_log_warn(state->log, "%s: follower delay:%ld target:%ld thr:%u, resync",
-					state->props.device, delay, target, state->threshold);
+			if (SPA_UNLIKELY(state->alsa_sync_warning)) {
+				spa_log_warn(state->log, "%s: follower delay:%ld target:%ld thr:%u, resync",
+						state->props.device, delay, target, state->threshold);
+				state->alsa_sync_warning = false;
+			} else
+				spa_log_info(state->log, "%s: follower delay:%ld target:%ld thr:%u, resync",
+						state->props.device, delay, target, state->threshold);
+
 			if (delay > target)
 				snd_pcm_rewind(state->hndl, delay - target);
 			else if (delay < target)
@@ -1903,8 +1945,6 @@ int spa_alsa_write(struct state *state)
 			delay = target;
 			state->alsa_sync = false;
 		}
-		if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, true)) < 0))
-			return res;
 	}
 
 	total_written = 0;
@@ -1928,56 +1968,37 @@ again:
 	written = 0;
 
 	while (!spa_list_is_empty(&state->ready) && to_write > 0) {
-		uint8_t *dst, *src;
 		size_t n_bytes, n_frames;
 		struct buffer *b;
 		struct spa_data *d;
-		uint32_t i, index, offs, avail, size, maxsize, l0, l1;
+		uint32_t i, offs, size;
 
 		b = spa_list_first(&state->ready, struct buffer, link);
 		d = b->buf->datas;
 
-		size = d[0].chunk->size;
-		maxsize = d[0].maxsize;
+		offs = d[0].chunk->offset + state->ready_offset;
+		size = d[0].chunk->size - state->ready_offset;
 
-		index = d[0].chunk->offset + state->ready_offset;
-		avail = size - state->ready_offset;
-		avail /= state->frame_size;
+		offs = SPA_MIN(offs, d[0].maxsize);
+		size = SPA_MIN(d[0].maxsize - offs, size);
 
-		n_frames = SPA_MIN(avail, to_write);
-		n_bytes = n_frames * state->frame_size;
-
-		offs = index % maxsize;
-		l0 = SPA_MIN(n_bytes, maxsize - offs);
-		l1 = n_bytes - l0;
+		n_frames = SPA_MIN(size / frame_size, to_write);
+		n_bytes = n_frames * frame_size;
 
 		if (SPA_LIKELY(state->use_mmap)) {
 			for (i = 0; i < b->buf->n_datas; i++) {
-				dst = SPA_PTROFF(my_areas[i].addr, off * state->frame_size, uint8_t);
-				src = d[i].data;
-
-				spa_memcpy(dst, src + offs, l0);
-				if (SPA_UNLIKELY(l1 > 0))
-					spa_memcpy(dst + l0, src, l1);
+				spa_memcpy(SPA_PTROFF(my_areas[i].addr, off * frame_size, void),
+						SPA_PTROFF(d[i].data, offs, void), n_bytes);
 			}
 		} else {
-			if (state->planar) {
-				void *bufs[b->buf->n_datas];
+			void *bufs[b->buf->n_datas];
+			for (i = 0; i < b->buf->n_datas; i++)
+				bufs[i] = SPA_PTROFF(d[i].data, offs, void);
 
-				for (i = 0; i < b->buf->n_datas; i++)
-					bufs[i] = SPA_PTROFF(d[i].data, offs, void);
-				snd_pcm_writen(hndl, bufs, l0 / state->frame_size);
-				if (SPA_UNLIKELY(l1 > 0)) {
-					for (i = 0; i < b->buf->n_datas; i++)
-						bufs[i] = d[i].data;
-					snd_pcm_writen(hndl, bufs, l1 / state->frame_size);
-				}
-			} else {
-				src = d[0].data;
-				snd_pcm_writei(hndl, src + offs, l0 / state->frame_size);
-				if (SPA_UNLIKELY(l1 > 0))
-					snd_pcm_writei(hndl, src, l1 / state->frame_size);
-			}
+			if (state->planar)
+				snd_pcm_writen(hndl, bufs, n_frames);
+			else
+				snd_pcm_writei(hndl, bufs[0], n_frames);
 		}
 
 		state->ready_offset += n_bytes;
@@ -2048,8 +2069,7 @@ push_frames(struct state *state,
 		spa_log_warn(state->log, "%s: no more buffers", state->props.device);
 		total_frames = frames;
 	} else {
-		uint8_t *src;
-		size_t n_bytes, left;
+		size_t n_bytes, left, frame_size = state->frame_size;
 		struct buffer *b;
 		struct spa_data *d;
 		uint32_t i, avail, l0, l1;
@@ -2065,23 +2085,26 @@ push_frames(struct state *state,
 
 		d = b->buf->datas;
 
-		avail = d[0].maxsize / state->frame_size;
+		avail = d[0].maxsize / frame_size;
 		total_frames = SPA_MIN(avail, frames);
-		n_bytes = total_frames * state->frame_size;
+		n_bytes = total_frames * frame_size;
 
 		if (my_areas) {
 			left = state->buffer_frames - offset;
-			l0 = SPA_MIN(n_bytes, left * state->frame_size);
+			l0 = SPA_MIN(n_bytes, left * frame_size);
 			l1 = n_bytes - l0;
 
 			for (i = 0; i < b->buf->n_datas; i++) {
-				src = SPA_PTROFF(my_areas[i].addr, offset * state->frame_size, uint8_t);
-				spa_memcpy(d[i].data, src, l0);
-				if (l1 > 0)
-					spa_memcpy(SPA_PTROFF(d[i].data, l0, void), my_areas[i].addr, l1);
+				spa_memcpy(d[i].data,
+						SPA_PTROFF(my_areas[i].addr, offset * frame_size, void),
+						l0);
+				if (SPA_UNLIKELY(l1 > 0))
+					spa_memcpy(SPA_PTROFF(d[i].data, l0, void),
+							my_areas[i].addr,
+							l1);
 				d[i].chunk->offset = 0;
 				d[i].chunk->size = n_bytes;
-				d[i].chunk->stride = state->frame_size;
+				d[i].chunk->stride = frame_size;
 			}
 		} else {
 			void *bufs[b->buf->n_datas];
@@ -2089,7 +2112,7 @@ push_frames(struct state *state,
 				bufs[i] = d[i].data;
 				d[i].chunk->offset = 0;
 				d[i].chunk->size = n_bytes;
-				d[i].chunk->stride = state->frame_size;
+				d[i].chunk->stride = frame_size;
 			}
 			if (state->planar) {
 				snd_pcm_readn(state->hndl, bufs, total_frames);
@@ -2131,9 +2154,18 @@ int spa_alsa_read(struct state *state)
 
 		avail = delay;
 
+		if (SPA_UNLIKELY((res = update_time(state, current_time, delay, target, true)) < 0))
+			return res;
+
 		if (state->alsa_sync) {
-			spa_log_warn(state->log, "%s: follower delay:%lu target:%lu thr:%u, resync",
-					state->props.device, delay, target, threshold);
+			if (SPA_UNLIKELY(state->alsa_sync_warning)) {
+				spa_log_warn(state->log, "%s: follower delay:%lu target:%lu thr:%u, resync",
+						state->props.device, delay, target, threshold);
+				state->alsa_sync_warning = false;
+			} else
+				spa_log_info(state->log, "%s: follower delay:%lu target:%lu thr:%u, resync",
+						state->props.device, delay, target, threshold);
+
 			if (delay < target)
 				max_read = target - delay;
 			else if (delay > target)
@@ -2142,9 +2174,6 @@ int spa_alsa_read(struct state *state)
 			state->alsa_sync = false;
 		}
 
-		if ((res = update_time(state, current_time, delay, target, true)) < 0)
-			return res;
-
 		if (avail < state->read_size)
 			max_read = 0;
 	}
@@ -2406,11 +2435,10 @@ int spa_alsa_start(struct state *state)
 	state->following = is_following(state);
 	setup_matching(state);
 
+	spa_dll_init(&state->dll);
 	state->threshold = (state->duration * state->rate + state->rate_denom-1) / state->rate_denom;
 	state->last_threshold = state->threshold;
-
-	spa_dll_init(&state->dll);
-	state->max_error = (256.0 * state->rate) / state->rate_denom;
+	state->max_error = SPA_MAX(256.0f, state->threshold / 2.0f);
 
 	spa_log_debug(state->log, "%p: start %d duration:%d rate:%d follower:%d match:%d resample:%d",
 			state, state->threshold, state->duration, state->rate_denom,
